# -*- coding: utf-8 -*-

'''
    Demo model in wxbtool package

    This model predict t850 3-days in the future
    it cost more memory and can be fitted into one P40 gpu at batch size 8
    the weighted rmse is ??? K
'''

import torch as th

from leibniz.nn.activation import CappingRelu
from leibniz.nn.net import resunet
from leibniz.unet.hyperbolic import HyperBottleneck

from wxbtool.specs.res5_625.t850rasp import Spec, Setting3d


class ResUNetModel(Spec):
    def __init__(self, setting):
        super().__init__(setting)
        self.name = 't850d3hg-rasp'

        self.resunet = resunet(setting.input_span * len(setting.vars_in) + self.constant_size + 2, 1,
                            spatial=(32, 64+2), layers=5, ratio=-2,
                            vblks=[9, 9, 9, 9, 9], hblks=[1, 1, 1, 1, 1],
                            scales=[-1, -1, -1, -1, -1], factors=[1, 1, 1, 1, 1],
                            block=HyperBottleneck, relu=CappingRelu(), final_normalized=False)

    def forward(self, **kwargs):
        batch_size = kwargs['temperature'].size()[0]
        self.update_da_status(batch_size)

        _, input = self.get_inputs(**kwargs)
        constant = self.get_augmented_constant(input)
        input = th.cat((input, constant), dim=1)
        input = th.cat((input[:, :, :, 63:64], input, input[:, :, :, 0:1]), dim=3)

        output = self.resunet(input)

        output = output[:, :, :, 1:65]
        return {
            't850': output
        }


setting = Setting3d()
model = ResUNetModel(setting)



# -*- coding: utf-8 -*-

'''
 A modeling spec for t850

 This spec follows basic settings and discussions in

   Improving Data‐Driven Global Weather Prediction Using Deep Convolutional Neural Networks on a Cubed Sphere
   by Jonathan A. Weyn, Dale R. Durran, Rich Caruana
   https://doi.org/10.1029/2020MS002109

'''

import torch as th
import torch.nn as nn
from wxbtool.nn.model import Base2d
from wxbtool.nn.setting import Setting
from wxbtool.norms.meanstd import *


mse = nn.MSELoss()


class SettingWeyn(Setting):
    def __init__(self):
        super().__init__()
        self.resolution = '5.625deg'    # The spatial resolution of the model

        self.levels = ['300', '500', '700', '850', '1000'] # Which vertical levels to choose
        self.height = len(self.levels)                     # How many vertical levels to choose

        # The name of variables to choose, for both input features and output
        self.vars = ['geopotential', 'toa_incident_solar_radiation', '2m_temperature', 'temperature']

        # The code of variables in input features
        self.vars_in = ['z500', 'z1000', 'tau', 't850', 't2m', 'tisr']
        # The code of variables in output
        self.vars_out = ['t850']

        # temporal scopes for train
        self.years_train = [
            1980, 1981, 1982, 1983, 1984, 1985, 1986, 1987, 1988, 1989,
            1990, 1991, 1992, 1993, 1994, 1995, 1996, 1997, 1998, 1999,
            2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009,
            2010, 2011, 2012, 2013, 2014,
        ]
        # temporal scopes for evaluation
        self.years_eval = [2015, 2016]
        # temporal scopes for test
        self.years_test = [2017, 2018]


class Setting3d(SettingWeyn):
    def __init__(self):
        super().__init__()
        self.step = 8                   # How many hours of a hourly step which all features in organized temporally
        self.input_span = 3             # How many hourly steps for an input
        self.pred_span = 1              # How many hourly steps for a prediction
        self.pred_shift = 72            # How many hours between the end of the input span and the beginning of prediction span


class SettingTest(SettingWeyn):
    def __init__(self):
        super().__init__()
        self.step = 8                   # How many hours of a hourly step which all features in organized temporally
        self.input_span = 3             # How many hourly steps for an input
        self.pred_span = 1              # How many hourly steps for a prediction
        self.pred_shift = 72            # How many hours between the end of the input span and the beginning of prediction span

        # temporal scopes for train
        self.years_train = [2013, 2014]
        # temporal scopes for evaluation
        self.years_eval = [2015, 2016]
        # temporal scopes for test
        self.years_test = [2017, 2018]


class Setting5d(SettingWeyn):
    def __init__(self):
        super().__init__()
        self.step = 8                   # How many hours of a hourly step which all features in organized temporally
        self.input_span = 3             # How many hourly steps for an input
        self.pred_span = 1              # How many hourly steps for a prediction
        self.pred_shift = 120           # How many hours between the end of the input span and the beginning of prediction span


class Spec(Base2d):
    def __init__(self, setting):
        super().__init__(setting)

        # following Weyn's schema from the above paper
        self.name = 't850_weyn'

    def get_inputs(self, **kwargs):
        z500 = norm_z500(kwargs['geopotential'].view(-1, self.setting.input_span, self.setting.height, 32, 64)[:, :, self.setting.levels.index('500')])
        z1000 = norm_z1000(kwargs['geopotential'].view(-1, self.setting.input_span, self.setting.height, 32, 64)[:, :, self.setting.levels.index('1000')])
        tau = norm_tau(kwargs['geopotential'].view(-1, self.setting.input_span, self.setting.height, 32, 64)[:, :, self.setting.levels.index('300')] - kwargs['geopotential'].view(-1, self.setting.input_span, self.setting.height, 32, 64)[:, :, self.setting.levels.index('700')])
        t850 = norm_t850(kwargs['temperature'].view(-1, self.setting.input_span, self.setting.height, 32, 64)[:, :, self.setting.levels.index('850')])
        t2m = norm_t2m(kwargs['2m_temperature'].view(-1, self.setting.input_span, 32, 64))
        tisr = norm_tisr(kwargs['toa_incident_solar_radiation'].view(-1, self.setting.input_span, 32, 64))

        z500 = self.augment_data(z500)
        z1000 = self.augment_data(z1000)
        tau = self.augment_data(tau)
        t850 = self.augment_data(t850)
        t2m = self.augment_data(t2m)
        tisr = self.augment_data(tisr)

        return {
            'z500': z500,
            'z1000': z1000,
            'tau': tau,
            't850': t850,
            't2m': t2m,
            'tisr': tisr,
        }, th.cat((
            z500, z1000,
            tau, t850,
            t2m, tisr,
        ), dim=1)

    def get_targets(self, **kwargs):
        t850 = kwargs['temperature'].view(-1, 1, self.setting.height, 32, 64)[:, :, self.setting.levels.index('850')]
        t850 = self.augment_data(t850)
        return {'t850': t850}, t850

    def get_results(self, **kwargs):
        t850 = denorm_t850(kwargs['t850'])
        return {'t850': t850}, t850

    def forward(self, **kwargs):
        raise NotImplementedError('Spec is abstract and can not be initialized')

    def lossfun(self, inputs, result, target):
        _, rst = self.get_results(**result)
        _, tgt = self.get_targets(**target)
        rst = self.weight * rst.view(-1, 1, 32, 64)
        tgt = self.weight * tgt.view(-1, 1, 32, 64)

        losst = mse(rst[:, 0], tgt[:, 0])

        return losst



from arghandler import *

from wxbtool.data.dsserver import main as dsmain
from wxbtool.nn.train import main as tnmain
from wxbtool.nn.test import main as ttmain


@subcmd
def help(parser, context, args):
    pass

@subcmd('dserve', help='start the dataset server')
def dserve(parser, context, args):
    parser.add_argument("-i", "--ip", type=str, default='127.0.0.1',
        help="the ip of the dataset serevr")
    parser.add_argument("-p", "--port", type=int, default=8088,
        help="the port of the dataset serevr")
    parser.add_argument("-w", "--workers", type=int, default=4,
        help="the number of workers")
    parser.add_argument("-m", "--module", type=str, default='wxbtool.specs.res5_625.t850weyn',
        help="module of a metrological model to load")
    parser.add_argument("-s", "--setting", type=str, default='Setting',
        help="setting for a metrological model spec")
    opt = parser.parse_args(args)

    dsmain(context, opt)


@subcmd('train', help='start training')
def train(parser, context, args):
    parser.add_argument("-g", "--gpu", type=str, default='0',
        help="index of gpu")
    parser.add_argument("-c", "--n_cpu", type=int, default=8,
        help="number of cpu threads to use during batch generation")
    parser.add_argument("-b", "--batch_size", type=int, default=64,
        help="size of the batches")
    parser.add_argument("-e", "--epoch", type=int, default=0,
        help="current epoch to start training from")
    parser.add_argument("-n", "--n_epochs", type=int, default=200,
        help="number of epochs of training")
    parser.add_argument("-m", "--module", type=str, default='wxbtool.zoo.unet.t850d3',
        help="module of the metrological model to load")
    parser.add_argument("-l", "--load", type=str, default='',
        help="dump file of the metrological model to load")
    parser.add_argument("-k", "--check", type=str, default='',
        help="checkpoint file to load")
    parser.add_argument("-r", "--rate", type=float, default=0.001,
        help="learning rate")
    parser.add_argument("-w", "--weightdecay", type=float, default=0.0,
        help="weight decay")
    parser.add_argument("-d", "--data", type=str, default='',
        help="url of the dataset server")
    opt = parser.parse_args(args)

    tnmain(context, opt)


@subcmd('test', help='start testing')
def test(parser, context, args):
    parser.add_argument("-g", "--gpu", type=str, default='0',
        help="index of gpu")
    parser.add_argument("-c", "--n_cpu", type=int, default=8,
        help="number of cpu threads to use during batch generation")
    parser.add_argument("-b", "--batch_size", type=int, default=64,
        help="size of the batches")
    parser.add_argument("-n", "--n_epochs", type=int, default=200,
        help="number of epochs of training")
    parser.add_argument("-m", "--module", type=str, default='wxbtool.zoo.unet.t850d3',
        help="module of the metrological model to load")
    parser.add_argument("-l", "--load", type=str, default='',
        help="dump file of the metrological model to load")
    parser.add_argument("-d", "--data", type=str, default='',
        help="url of the dataset server")
    opt = parser.parse_args(args)

    ttmain(context, opt)


def main():
    import sys

    handler = ArgumentHandler()
    if len(sys.argv) < 2:
        handler.run(['help'])
    else:
        handler.run(sys.argv[1:])



# -*- coding: utf-8 -*-

import os
import sys
import importlib
import logging
import resource
import arrow
import numpy as np
import torch as th

from pathlib import Path

from torch.optim.lr_scheduler import ReduceLROnPlateau
from torch.utils.data import DataLoader

from wxbtool.util.plotter import plot


rlimit = resource.getrlimit(resource.RLIMIT_NOFILE)
resource.setrlimit(resource.RLIMIT_NOFILE, (2048, rlimit[1]))

print('cudnn:', th.backends.cudnn.version())

np.core.arrayprint._line_width = 150
np.set_printoptions(linewidth=np.inf)

early_stopping = 5
best = np.inf
eval_by_epoch = np.inf
count = 0


embbed = False
scheduler = None


def train_model(opt, mdl, lr=0.001, wd=0.0, callback=None, model_path=None, logger=None):
    global scheduler
    optimizer = th.optim.Adam(mdl.parameters(), lr=lr, weight_decay=wd)
    scheduler = ReduceLROnPlateau(optimizer, 'min')

    try:
        if not embbed and opt.load != '':
            dump = th.load(opt.load, map_location='cpu')
            mdl.load_state_dict(dump)
    except ImportError as e:
        logger.exception(e)
        raise e

    if not embbed and th.cuda.is_available():
        mdl = mdl.cuda()

    def train(epoch, mdl):
        global eval_by_epoch
        mdl.train()
        dataloader = DataLoader(mdl.dataset_train, batch_size=opt.batch_size, shuffle=True, num_workers=opt.n_cpu)
        loss_per_epoch = 0.0
        for step, sample in enumerate(dataloader):
            inputs, targets = sample

            inputs = {
                v: th.as_tensor(inputs[v], dtype=th.float32) for v in mdl.setting.vars
            }
            targets = {
                v: th.as_tensor(targets[v], dtype=th.float32) for v in mdl.setting.vars
            }

            if not embbed and th.cuda.is_available():
                inputs = {
                    v: inputs[v].cuda() for v in mdl.setting.vars
                }
                targets = {
                    v: targets[v].cuda() for v in mdl.setting.vars
                }
                mdl.constant = mdl.constant.cuda()
                mdl.weight = mdl.weight.cuda()

            optimizer.zero_grad()
            results = mdl(*[], **inputs)
            loss = mdl.lossfun(inputs, results, targets)
            loss.backward()
            optimizer.step()

            logger.info(f'Epoch: {epoch + 1:03d} | Step: {step + 1:03d} | Loss: {loss.item()}')

            loss_per_epoch += loss.item() * list(results.values())[0].size()[0]

        logger.info(f'Epoch: {epoch + 1:03d} | Train Loss: {loss_per_epoch / mdl.train_size}')

        # evaluation
        loss_eval = evaluate(epoch)
        logger.info(f'Epoch: {epoch + 1:03d} | Eval loss: {loss_eval}')
        scheduler.step(loss_eval)
        eval_by_epoch = loss_eval

        if callback is not None:
            callback(epoch, loss_eval)

        global best, count
        if loss_eval.item() >= best:
            count += 1
        else:
            count = 0
            best = loss_eval

        if count == early_stopping:
            logger.info('early stopping reached, best loss is {:5f}'.format(best))

    def evaluate(epoch):
        mdl.eval()
        dataloader = DataLoader(mdl.dataset_eval, batch_size=opt.batch_size, shuffle=True, num_workers=opt.n_cpu)
        loss_per_epoch = 0.0
        rmse_per_epoch_t = 0.0
        for step, sample in enumerate(dataloader):
            inputs, targets = sample

            inputs = {
                v: th.as_tensor(inputs[v], dtype=th.float32) for v in mdl.setting.vars
            }
            targets = {
                v: th.as_tensor(targets[v], dtype=th.float32) for v in mdl.setting.vars
            }

            if th.cuda.is_available():
                inputs = {
                    v: inputs[v].cuda() for v in mdl.setting.vars
                }
                targets = {
                    v: targets[v].cuda() for v in mdl.setting.vars
                }
                mdl.constant = mdl.constant.cuda()
                mdl.weight = mdl.weight.cuda()

            with th.no_grad():
                results = mdl(*[], **inputs)
                loss = mdl.lossfun(inputs, results, targets)
                logger.info(f'Epoch: {epoch + 1:03d} | Step: {step + 1:03d} | Loss: {loss.item()}')
                loss_per_epoch += loss.item() * list(results.values())[0].size()[0]

                _, tgt = mdl.get_targets(**targets)
                _, rst = mdl.get_results(**results)
                tgt = tgt.detach().cpu().numpy().reshape(-1, 1, 32, 64)
                rst = rst.detach().cpu().numpy().reshape(-1, 1, 32, 64)
                rmse = np.sqrt(np.mean(mdl.weight.cpu().numpy() * (rst - tgt) * (rst - tgt)))
                logger.info(f'Epoch: {epoch + 1:03d} | Step: {step + 1:03d} | Loss: {loss.item()} | Temperature RMSE: {rmse}')
                rmse_per_epoch_t += np.nan_to_num(rmse * list(results.values())[0].size()[0])

        rmse_total = rmse_per_epoch_t / mdl.eval_size
        logger.info(f'Epoch: {epoch + 1:03d} | Eval Loss: {loss_per_epoch / mdl.eval_size}')
        logger.info(f'Epoch: {epoch + 1:03d} | Eval RMSE: {rmse_total}')

        vars_in, _ = mdl.get_inputs(**inputs)
        for bas, var in enumerate(mdl.setting.vars_in):
            for ix in range(mdl.setting.input_span):
                img = vars_in[var][0, ix].detach().cpu().numpy().reshape(32, 64)
                plot(var, open('%s_inp_%d.png' % (var, ix), mode='wb'), img)

        vars_fc, _ = mdl.get_results(**results)
        vars_tg, _ = mdl.get_targets(**targets)
        for bas, var in enumerate(mdl.setting.vars_out):
            fcst = vars_fc[var][0].detach().cpu().numpy().reshape(32, 64)
            tgrt = vars_tg[var][0].detach().cpu().numpy().reshape(32, 64)
            plot(var, open('%s_fcs.png' % var, mode='wb'), fcst)
            plot(var, open('%s_tgt.png' % var, mode='wb'), tgrt)

        th.save(mdl.state_dict(), model_path / f'm_rmse{rmse_total:0.8f}_epoch{epoch + 1:03d}.mdl')
        glb = list(model_path.glob('*.mdl'))
        if len(glb) > 6:
            for p in sorted(glb)[-3:]:
                os.unlink(p)

        th.save({
            'net': mdl.state_dict(),
            'optimizer': optimizer.state_dict(),
        }, model_path / f'z_epoch{epoch + 1:03d}.chk')
        glb = list(model_path.glob('*.chk'))
        if len(glb) > 1:
            os.unlink(sorted(glb)[0])

        return rmse_total

    try:
        for epoch in range(opt.n_epochs):
            train(epoch, mdl)
    except Exception as e:
        logger.exception(e)
        raise e


def main(context, opt):
    os.environ['CUDA_VISIBLE_DEVICES'] = opt.gpu
    try:
        mdm = importlib.import_module(opt.module, package=None)

        name = mdm.model.name
        time_str = arrow.now().format('YYYYMMDD_HHmmss')
        model_path = Path(f'./trains/{name}-{time_str}')
        model_path.mkdir(exist_ok=True, parents=True)
        log_file = model_path / Path('train.log')

        logging.basicConfig(level=logging.INFO, filename=log_file, filemode='w')
        logger = logging.getLogger()
        logger.setLevel(logging.INFO)
        handler = logging.StreamHandler(sys.stderr)
        handler.setFormatter(logging.Formatter("%(asctime)s %(levelname)s %(name)s: %(message)s"))
        logger.addHandler(handler)
        logger.info(str(opt))

        if mdm != None:
            if opt.data != '':
                mdm.model.load_dataset('train', 'client', url=opt.data)
            else:
                mdm.model.load_dataset('train', 'server')

            train_model(opt, mdm.model, lr=opt.rate, wd=opt.weightdecay, model_path=model_path, logger=logger)

        print('Training Finished!')
    except ImportError as e:
        print('failure when loading model')
        raise e



# -*- coding: utf-8 -*-

import os
import os.path as path
import hashlib
import requests
import logging
import json

import xarray as xr
import numpy as np

import msgpack
import msgpack_numpy as m
m.patch()


from itertools import product
from torch.utils.data import Dataset

from wxbtool.data.variables import codes, vars2d, vars3d

logger = logging.getLogger()

all_levels = ['50', '100', '150', '200', '250', '300', '400', '500', '600', '700', '850', '925', '1000']


class WxDataset(Dataset):
    def __init__(self, root, resolution, years, vars, levels, step=1, input_span=2, pred_shift=24, pred_span=1):
        self.root = root
        self.resolution = resolution
        self.input_span = input_span
        self.step = step
        self.pred_shift = pred_shift
        self.pred_span = pred_span
        self.years = years
        self.vars = vars
        self.levels = levels
        self.inputs = {}
        self.targets = {}

        if resolution == '5.625deg':
            self.height = 13
            self.width = 32
            self.length = 64

        code = '%s:%s:%s:%s:%s:%s:%s:%s' % (resolution, years, vars, levels, step, input_span, pred_shift, pred_span)
        hashstr = hashlib.md5(code.encode('utf-8')).hexdigest()
        self.hashcode = hashstr

        dumpdir = path.abspath('%s/.cache/%s' % (self.root, hashstr))
        if not path.exists(dumpdir):
            self.load()
            os.makedirs(dumpdir)
            self.dump(dumpdir)

        self.memmap(dumpdir)

    def load(self):
        levels_selector = []
        for l in self.levels:
            levels_selector.append(all_levels.index(l))
        selector = np.array(levels_selector, dtype=np.int)

        size, dti, dto = self.init_holders(self.vars)
        for var, yr in product(self.vars, self.years):
            if var in vars3d:
                length, chi, cho = self.load_3ddata(yr, var, selector)
            elif var in vars2d:
                length, chi, cho = self.load_2ddata(yr, var)
            else:
                raise ValueError('variable %s dose not supported!' % var)
            size = size + length
            dti[var].append(chi)
            dto[var].append(cho)

        self.size = size // len(self.vars)
        for v in self.vars:
            input = np.concatenate(tuple(dti[v]), axis=0)
            target = np.concatenate(tuple(dto[v]), axis=0)
            dti[v] = None
            dto[v] = None

            self.inputs[v] = input
            self.targets[v] = target
        logger.info('total %s items loaded!', self.size)

    def init_holders(self, vars):
        return 0, {k: [] for k in vars}, {k: [] for k in vars}

    def load_2ddata(self, year, var):
        data_path = '%s/%s/%s_%d_%s.nc' % (self.root, var, var, year, self.resolution)
        ds = xr.open_dataset(data_path)
        ds = ds.transpose('time', 'lat', 'lon')
        dt = np.array(ds[codes[var]].data, dtype=np.float32)
        logger.info('%s[%d]: %s', var, year, str(dt.shape))

        length = 365 * 24 - (self.input_span * self.step + self.pred_span * self.step + self.pred_shift)
        dti, dto = (
            np.zeros([length, self.input_span, self.width, self.length], dtype=np.float32),
            np.zeros([length, self.pred_span, self.width, self.length], dtype=np.float32)
        )

        for ix in range(0, length):
            pt = ix
            for jx in range(self.input_span):
                dti[pt, jx, :, :] = dt[ix + jx * self.step, :, :]
            for kx in range(self.pred_span):
                dto[pt, kx, :, :] = dt[ix + self.pred_shift + kx * self.step, :, :]

        ds.close()

        return length, dti, dto

    def load_3ddata(self, year, var, selector):
        data_path = '%s/%s/%s_%d_%s.nc' % (self.root, var, var, year, self.resolution)
        ds = xr.open_dataset(data_path)
        ds = ds.transpose('time', 'level', 'lat', 'lon')
        dt = np.array(ds[codes[var]].data, dtype=np.float32)
        logger.info('%s[%d]: %s', var, year, str(dt.shape))

        height = selector.shape[0]

        length = 365 * 24 - (self.input_span * self.step + self.pred_span * self.step + self.pred_shift)
        dti, dto = (
            np.zeros([length, self.input_span, height, self.width, self.length], dtype=np.float32),
            np.zeros([length, self.pred_span, height, self.width, self.length], dtype=np.float32)
        )

        for ix in range(0, length):
            pt = ix
            for jx in range(self.input_span):
                dti[pt, jx, :, :, :] = dt[ix + jx * self.step, selector, :, :]
            for kx in range(self.pred_span):
                dto[pt, kx, :, :, :] = dt[ix + self.pred_shift + kx * self.step, selector, :, :]

        ds.close()

        return length, dti, dto

    def dump(self, dumpdir):
        shapes = {
            'inputs': {},
            'targets': {},
        }

        for var in self.vars:
            input_dump = '%s/input_%s.npy' % (dumpdir, var)
            target_dump = '%s/target_%s.npy' % (dumpdir, var)

            shapes['inputs'][var] = self.inputs[var].shape
            np.save(input_dump, self.inputs[var])
            del self.inputs[var]

            shapes['targets'][var] = self.targets[var].shape
            np.save(target_dump, self.targets[var])
            del self.targets[var]

        with open('%s/shapes.json' % dumpdir, mode='w') as fp:
            json.dump(shapes, fp)

    def memmap(self, dumpdir):
        with open('%s/shapes.json' % dumpdir) as fp:
            shapes = json.load(fp)

        for var in self.vars:
            input_dump = '%s/input_%s.npy' % (dumpdir, var)
            target_dump = '%s/target_%s.npy' % (dumpdir, var)

            shape = shapes['inputs'][var]
            total_size = np.prod(shape)
            input = np.memmap(input_dump, dtype=np.float32, mode='r')
            shift = input.shape[0] - total_size
            self.inputs[var] = np.reshape(input[shift:], shape)

            shape = shapes['targets'][var]
            total_size = np.prod(shape)
            target = np.memmap(target_dump, dtype=np.float32, mode='r')
            shift = target.shape[0] - total_size
            self.targets[var] = np.reshape(target[shift:], shape)

    def __len__(self):
        return self.inputs[self.vars[0]].shape[0]

    def __getitem__(self, item):
        inputs, targets = {}, {}
        for var in self.vars:
            inputs.update({var: self.inputs[var][item:item+1]})
            targets.update({var: self.targets[var][item:item+1]})
        return inputs, targets


class WxDatasetClient(Dataset):
    def __init__(self, url, phase, resolution, years, vars, levels, step=1, input_span=2, pred_shift=24, pred_span=1):
        self.url = url
        self.phase = phase

        code = '%s:%s:%s:%s:%s:%s:%s:%s' % (resolution, years, vars, levels, step, input_span, pred_shift, pred_span)
        self.hashcode = hashlib.md5(code.encode('utf-8')).hexdigest()

    def __len__(self):
        url = '%s/%s/%s'% (self.url, self.hashcode, self.phase)
        r = requests.get(url)
        if r.status_code != 200:
            raise Exception('http error %s: %s' % (r.status_code, r.text))

        data = msgpack.loads(r.content)

        return data['size']

    def __getitem__(self, item):
        url = '%s/%s/%s/%d'% (self.url, self.hashcode, self.phase, item)
        r = requests.get(url)
        if r.status_code != 200:
            raise Exception('http error %s: %s' % (r.status_code, r.text))

        data = msgpack.loads(r.content)
        for key, val in data.items():
            for var, blk in val.items():
                val[var] = np.copy(blk)

        return data['inputs'], data['targets']



# -*- coding: utf-8 -*-

'''
 A modeling spec for t850

 This spec follows basic settings and discussions in

   Data-driven medium-range weather prediction with a Resnet pretrained on climate simulations:  A new model for WeatherBench
   by Stephan Rasp, Nils Thuerey
   https://arxiv.org/pdf/2008.08626.pdf

'''

import torch as th
import torch.nn as nn
from wxbtool.nn.model import Base2d
from wxbtool.nn.setting import Setting
from wxbtool.data.variables import vars3d, code2var, split_name
from wxbtool.norms.meanstd import normalizors, denorm_t850


mse = nn.MSELoss()


class SettingRasp(Setting):
    def __init__(self):
        super().__init__()
        self.resolution = '5.625deg'    # The spatial resolution of the model

        # Which vertical levels to choose
        self.levels = ['50', '250', '500', '600', '700', '850', '925']
        # How many vertical levels to choose
        self.height = len(self.levels)

        # The name of variables to choose, for both input features and output
        self.vars = ['geopotential', 'temperature', 'specific_humidity', 'u_component_of_wind', 'v_component_of_wind',
                     '2m_temperature', 'total_precipitation', 'toa_incident_solar_radiation']

        # The code of variables in input features
        self.vars_in = ['z50', 'z250', 'z500', 'z600', 'z700', 'z850', 'z925',
                        't50', 't250', 't500', 't600', 't700', 't850', 't925',
                        'q50', 'q250', 'q500', 'q600', 'q700', 'q850', 'q925',
                        'u50', 'u250', 'u500', 'u600', 'u700', 'u850', 'u925',
                        'v50', 'v250', 'v500', 'v600', 'v700', 'v850', 'v925',
                        't2m', 'tp', 'tisr']

        # The code of variables in output
        self.vars_out = ['t850']

        # temporal scopes for train
        self.years_train = [
            1980, 1981, 1982, 1983, 1984, 1985, 1986, 1987, 1988, 1989,
            1990, 1991, 1992, 1993, 1994, 1995, 1996, 1997, 1998, 1999,
            2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009,
            2010, 2011, 2012, 2013, 2014,
        ]
        # temporal scopes for evaluation
        self.years_eval = [2015, 2016]
        # temporal scopes for test
        self.years_test = [2017, 2018]


class Setting3d(SettingRasp):
    def __init__(self):
        super().__init__()
        self.step = 8                   # How many hours of a hourly step which all features in organized temporally
        self.input_span = 3             # How many hourly steps for an input
        self.pred_span = 1              # How many hourly steps for a prediction
        self.pred_shift = 72            # How many hours between the end of the input span and the beginning of prediction span


class Setting5d(SettingRasp):
    def __init__(self):
        super().__init__()
        self.step = 8                   # How many hours of a hourly step which all features in organized temporally
        self.input_span = 3             # How many hourly steps for an input
        self.pred_span = 1              # How many hourly steps for a prediction
        self.pred_shift = 120           # How many hours between the end of the input span and the beginning of prediction span


class Spec(Base2d):
    def __init__(self, setting):
        super().__init__(setting)

        # following Rasp's schema from the above paper
        self.name = 't850_rasp'

    def get_inputs(self, **kwargs):
        vdic, vlst = {}, []
        for nm in self.setting.vars_in:
            c, l = split_name(nm)
            v = code2var[c]
            if v in vars3d:
                d = kwargs[v].view(-1, self.setting.input_span, self.setting.height, 32, 64)[:, :, self.setting.levels.index(l)]
            else:
                d = kwargs[v].view(-1, self.setting.input_span, 32, 64)
            d = normalizors[nm](d)
            d = self.augment_data(d)
            vdic[nm] = d
            vlst.append(d)

        return vdic, th.cat(vlst, dim=1)

    def get_targets(self, **kwargs):
        t850 = kwargs['temperature'].view(-1, 1, self.setting.height, 32, 64)[:, :, self.setting.levels.index('850')]
        t850 = self.augment_data(t850)
        return {'t850': t850}, t850

    def get_results(self, **kwargs):
        t850 = denorm_t850(kwargs['t850'])
        return {'t850': t850}, t850

    def forward(self, **kwargs):
        raise NotImplementedError('Spec is abstract and can not be initialized')

    def lossfun(self, inputs, result, target):
        _, rst = self.get_results(**result)
        _, tgt = self.get_targets(**target)
        rst = self.weight * rst.view(-1, 1, 32, 64)
        tgt = self.weight * tgt.view(-1, 1, 32, 64)

        losst = mse(rst[:, 0], tgt[:, 0])

        return losst
